from enum import Enum

"""
These are the correct values on the IDT website as of time of writing
Please check https://eu.idtdna.com/site/order/oligoentry to confirm
"""


class PlateSize(Enum):
    WELL96 = "96"
    WELL384 = "384"


class PlateSplitBy(Enum):
    POOL = "pool"
    REF = "ref"
    REF_POOL = "ref_pool"
    NONE = "none"


class PlateFillBy(Enum):
    ROWS = "rows"
    COLS = "cols"


class TubeScale(Enum):
    NM25 = "25nm"
    NM100 = "100nm"
    NM250 = "250nm"
    UM1 = "1um"
    UM5 = "5um"
    UM10 = "10um"


class TubePurification(Enum):
    STD = "STD"
    PAGE = "PAGE"
    HPLC = "HPLC"
    IEHPLC = "IEHPLC"
    RNASE = "RNASE"
    DUALHPLC = "DUALHPLC"
    PAGEHPLC = "PAGEHPLC"



import pathlib
import unittest
import tempfile
import xlsxwriter
from primalbedtools.scheme import Scheme

from bed2idt.main import plates_go
from bed2idt.config import PlateSplitBy, PlateFillBy, PlateSize

import pandas as pd

BED_PATH = pathlib.Path(__file__).parent / "test_input/primer.bed"


class TestPlateWriter(unittest.TestCase):
    def setUp(self) -> None:
        self.scheme = Scheme.from_file(str(BED_PATH))
        return super().setUp()

    def test_write_96(self):
        """Write a 96 well plate"""

        with tempfile.TemporaryDirectory(
            dir="tests", suffix="-96well-plate"
        ) as tempdir:
            tempdir_path = pathlib.Path(tempdir)
            output = tempdir_path / "output.xlsx"
            workbook = xlsxwriter.Workbook(output)

            # Write the plate
            plates_go(
                self.scheme,
                workbook,
                PlateSplitBy.REF,
                PlateFillBy.ROWS,
                "test",
                False,
                PlateSize.WELL96,
            )

            # Check the plate is as expected
            df = pd.read_excel(output, sheet_name=None)

            pn_to_seq = {bl.primername: bl.sequence for bl in self.scheme.bedlines}
            found_pn = set()

            sheet_names = []
            # Check Sequences match
            for _sheetname, sheet in df.items():
                # Check structure
                self.assertEqual(
                    ["Well Position", "Name", "Sequence"], sheet.columns.to_list()
                )
                # Check primernames are present
                for row in sheet.itertuples():
                    # check correct seq is provided
                    self.assertEqual(pn_to_seq[row.Name], row.Sequence)  # type: ignore
                    found_pn.add(row.Name)

                sheet_names.append(_sheetname)

            # Check all primers are present
            self.assertSetEqual(found_pn, set(pn_to_seq.keys()))
            # Check all sheet names are present
            self.assertEqual(sheet_names, ["test_MN908947.3.0", "test_MN908947.3.1"])

    def test_write_96_split_ref_pool(self):
        """Write a 96 well plate"""

        with tempfile.TemporaryDirectory(
            dir="tests", suffix="-96well-plate"
        ) as tempdir:
            tempdir_path = pathlib.Path(tempdir)
            output = tempdir_path / "output.xlsx"
            workbook = xlsxwriter.Workbook(output)

            # Write the plate
            plates_go(
                self.scheme,
                workbook,
                PlateSplitBy.REF_POOL,
                PlateFillBy.ROWS,
                "test",
                False,
                PlateSize.WELL96,
            )

            # Check the plate is as expected
            df = pd.read_excel(output, sheet_name=None)

            pn_to_seq = {bl.primername: bl.sequence for bl in self.scheme.bedlines}
            found_pn = set()

            sheet_names = []
            # Check Sequences match
            for _sheetname, sheet in df.items():
                # Check structure
                self.assertEqual(
                    ["Well Position", "Name", "Sequence"], sheet.columns.to_list()
                )
                # Check primernames are present
                for row in sheet.itertuples():
                    # check correct seq is provided
                    self.assertEqual(pn_to_seq[row.Name], row.Sequence)  # type: ignore
                    found_pn.add(row.Name)

                sheet_names.append(_sheetname)

            # Check all primers are present
            self.assertSetEqual(found_pn, set(pn_to_seq.keys()))
            # Check all sheet names are present
            self.assertEqual(
                sheet_names, ["test_MN908947.3_1.0", "test_MN908947.3_2.0"]
            )

    def test_write_384(self):
        """Write a 384 well plate"""

        with tempfile.TemporaryDirectory(
            dir="tests", suffix="-384well-plate"
        ) as tempdir:
            tempdir_path = pathlib.Path(tempdir)
            output = tempdir_path / "output.xlsx"
            workbook = xlsxwriter.Workbook(output)

            # Write the plate
            plates_go(
                self.scheme,
                workbook,
                PlateSplitBy.POOL,
                PlateFillBy.ROWS,
                "test",
                False,
                PlateSize.WELL384,
            )

            # Check the plate is as expected
            df = pd.read_excel(output, sheet_name=None)

            pn_to_seq = {bl.primername: bl.sequence for bl in self.scheme.bedlines}
            found_pn = set()

            sheet_names = []
            # Check Sequences match
            for _sheetname, sheet in df.items():
                # Check structure
                self.assertEqual(
                    ["Well Position", "Name", "Sequence"], sheet.columns.to_list()
                )
                # Check primernames are present
                for row in sheet.itertuples():
                    # check correct seq is provided
                    self.assertEqual(pn_to_seq[row.Name], row.Sequence)  # type: ignore
                    found_pn.add(row.Name)
                sheet_names.append(_sheetname)

            # Check all primers are present
            self.assertSetEqual(found_pn, set(pn_to_seq.keys()))

            # Check all sheet names are present
            self.assertEqual(sheet_names, ["test_1.0", "test_2.0"])


if __name__ == "__main__":
    unittest.main()



import pathlib
import unittest

from bed2idt.main import append_xlsx, chunks


class TestFuncs(unittest.TestCase):
    def test_chunks(self):
        lst = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        n = 3
        result = list(chunks(lst, n))
        expected = [[1, 2, 3], [4, 5, 6], [7, 8, 9], [10]]
        self.assertEqual(result, expected)

    def test_append_xlsx(self):
        valid_path = pathlib.Path("fakepath.xlsx")
        invalid_path = pathlib.Path("fakepath")

        # Test valid path
        result_valid_path = append_xlsx(valid_path)
        self.assertEqual(result_valid_path, valid_path)

        # Test invalid path
        result_invalid_path = append_xlsx(invalid_path)
        self.assertEqual(result_invalid_path, invalid_path.with_suffix(".xlsx"))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) 2025, Chris Kent
# All rights reserved.
#
# This source code is licensed under the BSD-3-Clause license found in the
# LICENSE file in the root directory of this source tree.

import pathlib
import random
from importlib.metadata import version

import typer
import xlsxwriter
from primalbedtools.bedfiles import BedLine, group_by_chrom, group_by_pool
from primalbedtools.scheme import Scheme
from typing_extensions import Annotated

from bed2idt.config import (
    PlateFillBy,
    PlateSize,
    PlateSplitBy,
    TubePurification,
    TubeScale,
)


# Create the typer app
app = typer.Typer(
    name="bed2idt", no_args_is_help=True, pretty_exceptions_show_locals=False
)

SEED = 2025_08_20

plate_letters = [
    "A",
    "B",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "I",
    "J",
    "K",
    "L",
    "M",
    "N",
    "O",
    "P",
]
plate_numbers = [*range(1, 25)]


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def create_plates(
    primer_list: list[list[BedLine]],
    workbook,
    sheet_name: str,
    by_rows: bool,
    size: PlateSize,
):
    for index, primer_sublist in enumerate(primer_list):
        # Create all the indexes in a generator
        if size == PlateSize.WELL96:
            letters = plate_letters[:8]
            numb = plate_numbers[:12]
        else:
            letters = plate_letters[:16]
            numb = plate_numbers[:24]

        # Ensure the plate is filled up correctly
        if by_rows:
            indexes = ((letter, str(number)) for letter in letters for number in numb)
        else:
            indexes = ((letter, str(number)) for number in numb for letter in letters)

        # Create the sheet
        worksheet = workbook.add_worksheet(f"{sheet_name}.{index}")

        headings = ["Well Position", "Name", "Sequence"]
        for index, head in enumerate(headings):
            worksheet.write(0, index, head)
        row = 1
        content = [
            ("".join(i), p.primername, p.sequence)
            for i, p in zip(indexes, primer_sublist)
        ]

        for primer in content:
            for col, data in enumerate(primer):
                worksheet.write(row, col, data)
            row += 1


def plates_go(
    scheme: Scheme,
    workbook,
    splitby: PlateSplitBy,
    fillby: PlateFillBy,
    plateprefix: str,
    randomise: bool,
    size: PlateSize,
):
    if randomise:
        random.Random(SEED).shuffle(scheme.bedlines)

    splits: dict[str, list[BedLine]] = dict()

    # Ensure primers are split by pool
    if splitby == PlateSplitBy.POOL:
        splits = {str(k): v for k, v in group_by_pool(scheme.bedlines).items()}

    # Don't Split plates by primers
    elif splitby == PlateSplitBy.NONE:
        splits = {"": scheme.bedlines}

    # Split primers by the ref genome
    elif splitby == PlateSplitBy.REF:
        splits = group_by_chrom(scheme.bedlines)

    elif splitby == PlateSplitBy.REF_POOL:
        ref_splits = group_by_chrom(scheme.bedlines)

        for ref, ref_bedlines in ref_splits.items():
            ref_pool_splits = {
                str(k): v for k, v in group_by_pool(ref_bedlines).items()
            }
            for pool, ref_pool_bedlines in ref_pool_splits.items():
                splits[f"{ref}_{pool}"] = ref_pool_bedlines

    else:
        raise typer.BadParameter("Please select a valid option for --splitby")

    # make sure no pool are more than 96 or 384 primers
    split_plates = {
        k: list(chunks(v, 96 if size == PlateSize.WELL96 else 384))
        for k, v in splits.items()
    }  # type: ignore

    for split, plates in split_plates.items():
        if plates:
            create_plates(
                plates,  # type: ignore
                workbook,
                sheet_name=f"{plateprefix}_{split}" if split else f"{plateprefix}",
                by_rows=fillby == PlateFillBy.ROWS,
                size=size,
            )

    workbook.close()


def tubes_go(
    scheme: Scheme, workbook, scale: TubeScale, purification: TubePurification
):
    worksheet = workbook.add_worksheet()
    headings = ["Name", "Sequence", "Scale", "Purification"]

    # Write the headings
    for index, head in enumerate(headings):
        worksheet.write(0, index, head)
    row = 1

    # Generate the things to write
    content = [
        (x.primername, x.sequence, scale.value, purification.value)
        for x in scheme.bedlines
    ]

    # Write each line
    for primer in content:
        for col, data in enumerate(primer):
            worksheet.write(row, col, data)

        row += 1

    workbook.close()


def append_xlsx(path: pathlib.Path):
    """
    Append to an xlsx file path
    """
    if path.suffix != ".xlsx":
        return path.with_suffix(".xlsx")
    return path


def typer_callback_version(value: bool):
    if value:
        typer.echo(f"bed2idt version: {version('bed2idt')}")
        raise typer.Exit()


@app.callback()
def common(
    value: Annotated[bool, typer.Option] = typer.Option(
        False, "--version", callback=typer_callback_version
    ),
):
    pass


@app.command(no_args_is_help=True)
def plates(
    bedfile: Annotated[
        pathlib.Path,
        typer.Argument(help="The path to the bed file", readable=True),
    ],
    plateprefix: Annotated[
        str,
        typer.Option(
            help="The prefix used in naming sheets in the excel file",
            show_default=False,
        ),
    ],
    output: Annotated[
        pathlib.Path,
        typer.Option(
            help="The output location of the file. Defaults to output.xlsx",
            writable=True,
            callback=append_xlsx,
            dir_okay=False,
        ),
    ] = pathlib.Path("output.xlsx"),
    splitby: Annotated[
        PlateSplitBy,
        typer.Option(help="Should the primers be split across different plate"),
    ] = PlateSplitBy.POOL.value,  # type: ignore
    fillby: Annotated[
        PlateFillBy, typer.Option(help="How should the plate be filled")
    ] = PlateFillBy.COLS.value,  # type: ignore
    force: Annotated[
        bool, typer.Option(help="Override the output directory", show_default=False)
    ] = False,
    randomise: Annotated[
        bool,
        typer.Option(help="Randomise the order of primers within a plate"),
    ] = False,
    platesize: Annotated[
        PlateSize,
        typer.Option(help="The size of the plate to use"),
    ] = PlateSize.WELL96.value,  # type: ignore
):
    # Check the outpath
    if output.exists() and not force:
        raise typer.BadParameter(
            f"File exists at {output.absolute()}, add --force to overwrite"
        )
    # Read in the primers
    scheme = Scheme.from_file(str(bedfile))

    # Create the workbook
    workbook = xlsxwriter.Workbook(output)

    # Create the plates
    plates_go(scheme, workbook, splitby, fillby, plateprefix, randomise, platesize)


@app.command(no_args_is_help=True)
def tubes(
    bedfile: Annotated[
        pathlib.Path, typer.Argument(help="The path to the bed file", readable=True)
    ],
    output: Annotated[
        pathlib.Path,
        typer.Option(
            help="The output location of the file. Defaults to output.xlsx",
            writable=True,
            callback=append_xlsx,
            dir_okay=False,
        ),
    ] = pathlib.Path("output.xlsx"),
    scale: Annotated[
        TubeScale, typer.Option(help="The concentration of the primers")
    ] = TubeScale.NM25.value,  # type: ignore
    purification: Annotated[
        TubePurification, typer.Option(help="The purification of the primers")
    ] = TubePurification.STD.value,  # type: ignore
    force: Annotated[
        bool, typer.Option(help="Override the output directory", show_default=False)
    ] = False,
):
    # Check the outpath
    if output.exists() and not force:
        raise typer.BadParameter(
            f"File exists at {output.absolute()}, add --force to overwrite"
        )

    # Read in the primers
    scheme = Scheme.from_file(str(bedfile))
    # Create the workbook
    workbook = xlsxwriter.Workbook(output)

    # Create the tubes
    tubes_go(scheme, workbook, scale, purification)


if __name__ == "__main__":
    app()



import pathlib

from typer.testing import CliRunner
from importlib.metadata import version

from bed2idt.main import app

runner = CliRunner()

# This uses pytest rather than unittest
# Run using: poetry run pytest

BEDFILE_INPUT = pathlib.Path("tests/test_input/primer.bed")
EXISTING_OUTPUT = pathlib.Path("tests/test_output/output.xlsx")
NON_EXISTING_OUTPUT = pathlib.Path("tests/test_output/output2.xlsx")

# Ensure NON_EXISTING_OUTPUT does not exist
if NON_EXISTING_OUTPUT.exists():
    NON_EXISTING_OUTPUT.unlink()


# Test the app can run with the version flag
def test_app_version():
    result = runner.invoke(app, ["--version"])
    assert result.exit_code == 0
    assert version("bed2idt") in result.stdout


def test_plate_generate_exists():
    result = runner.invoke(
        app,
        [
            "plates",
            "--plateprefix",
            "test",
            "--output",
            str(EXISTING_OUTPUT.absolute()),
            str(BEDFILE_INPUT.absolute()),
        ],
    )
    assert result.exit_code == 2  # Should error as the output file already exists


def test_plate_generate_okay():
    result = runner.invoke(
        app,
        [
            "plates",
            "--plateprefix",
            "test",
            "--output",
            str(NON_EXISTING_OUTPUT.absolute()),
            str(BEDFILE_INPUT.absolute()),
        ],
    )
    assert result.exit_code == 0  # Be okay as the output file does not exist
    assert NON_EXISTING_OUTPUT.exists()  # Check the output file was created

    # Clean up the output file
    NON_EXISTING_OUTPUT.unlink()



import pathlib
import unittest
import tempfile
import xlsxwriter
from primalbedtools.scheme import Scheme

from bed2idt.main import tubes_go
from bed2idt.config import TubePurification, TubeScale
import pandas as pd

BED_PATH = pathlib.Path(__file__).parent / "test_input/primer.bed"


class TestTubeWriter(unittest.TestCase):
    def setUp(self) -> None:
        self.scheme = Scheme.from_file(str(BED_PATH))
        return super().setUp()

    def test_write_tubes(self):
        """Write tubes"""

        with tempfile.TemporaryDirectory(dir="tests", suffix="-tube") as tempdir:
            tempdir_path = pathlib.Path(tempdir)
            output = tempdir_path / "output.xlsx"
            workbook = xlsxwriter.Workbook(output)

            scale = TubeScale.NM25
            purification = TubePurification.STD

            # Write the plate
            tubes_go(self.scheme, workbook, scale, purification)

            # Check the plate is as expected
            df = pd.read_excel(output, sheet_name=None)

            pn_to_seq = {bl.primername: bl.sequence for bl in self.scheme.bedlines}
            found_pn = set()

            # Check Sequences match
            for _sheetname, sheet in df.items():
                # Check structure
                self.assertEqual(
                    ["Name", "Sequence", "Scale", "Purification"],
                    sheet.columns.to_list(),
                )
                # Check primernames are present
                for row in sheet.itertuples():
                    # check correct seq is provided
                    self.assertEqual(pn_to_seq[row.Name], row.Sequence)  # type: ignore
                    found_pn.add(row.Name)

                    # Check the Scale and purification
                    self.assertEqual(row.Scale, scale.value)
                    self.assertEqual(row.Purification, purification.value)

            # Check all primers are present
            self.assertSetEqual(found_pn, set(pn_to_seq.keys()))


if __name__ == "__main__":
    unittest.main()
